#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
import netaddr
import netifaces
import os
import socket
from argparse import ArgumentParser
from asyncio import async, coroutine, get_event_loop, wait, sleep
from lockfile import LockFile
from netaddr import IPNetwork
from nsenter import Namespace
from pathlib import Path
import sys
import pequod_agent
from pequod_agent.docker_manager import DockerManager
from pequod_agent.cluster import Cluster
from pequod_agent.constants.net import BRIDGE_NAME
from pequod_agent.constants.path import NETNS_PATH, CACHE_PATH, DATA_PATH, PERSISTENCE_PATH, INSTANCE_DATA_PATH, \
    RESOLV_CONF_PATH, LOCK_PATH
from pequod_agent.drydock.networking import setup_routes
from pequod_agent.errors import ConfigurationError, StartupError, PreflightError
from pequod_agent.inspect import inspect
from pequod_agent.losetup import get_loop_devices
from pequod_agent.models.instance import StartingInstance, Instance
from pequod_agent.models.node_state import NodeState
from pequod_agent.utilities.colorlog import ColorLoggingHandler
from pequod_agent.utilities.mount import get_mounts
from pequod_agent.networking.ip import wait_for_dad, get_host_ip
from pequod_agent.networking.ippool import IPPool
from pequod_agent.networking.iptables import ensure
from pequod_agent.networking.routing.haproxy import HAProxyManager
from pequod_agent.starter import Starter
from pequod_agent.utilities.async import run_async
from pequod_agent.utilities.system import execute, reboot, remove_users
from setproctitle import setproctitle
from shutil import rmtree
from time import time


MAIN_LOOP_INTERVAL = 1.0
CLEANUP_INTERVAL = 15

CORE_SEEDS_ARG = 'core-seeds='


class Agent:
    def __init__(self, cluster: Cluster, node_state: NodeState, test_mode=False, expose_services=False):
        self.lb_updater = cluster.lb_updater
        self.core_manager = cluster.core_manager
        self.host = socket.getfqdn()
        self.ctrl = cluster.controller
        try:
            self.ctrl.service.on_master_change = self.register
        except AttributeError:
            # FIXME: catching AttributeError is dangerous as it might hide bugs!
            # Bootstrap controller doesn't have a service
            pass
        self.registry = cluster.registry
        self.cluster_state = {}
        self.docker = DockerManager()
        # FIXME: really hacky setting the credentials here,
        # but the singleton approach makes things more complicated
        # (you never know where the instance is really instantiated)
        self.docker.credentials = cluster.credentials
        self.node_state = node_state
        self.log = logging.getLogger('pequod.agent')
        self.test_mode = test_mode
        self.expose_services = expose_services
        self.haproxy_manager = HAProxyManager()
        self.my_location = None
        self._last_haproxy_loop = time()
        self._last_cleanup_loop = time()
        self._last_iptable_loop = time()

    def setproctitle(self, tpl, *args):
        setproctitle('pequod-agent: {}'.format(tpl.format(*args)))

    def cleanup_containers(self):
        now = time()
        for container in self.docker.all_containers:
            if not container['Status'].startswith('Up'):
                created = container['Created']
                seconds_ago = now - created
                # we keep the Docker container around for some time to help debugging (i.e. check logs)
                if seconds_ago > 3600 * 24:
                    cid = container['Id']
                    self.log.info('Deleting %s.. (%s)', cid[:12], container['Status'])
                    self.docker.remove_container(container)
                else:
                    instance_name = Instance.get_name_from_container_name(container['Names'][0])
                    instance_path = INSTANCE_DATA_PATH / instance_name
                    if instance_path.is_dir():
                        self.log.info('Deleting instance %s..', instance_name)
                        loop_devices = {}
                        for loop in get_loop_devices():
                            if instance_path in Path(loop.file).parents:
                                loop_devices[loop.device] = loop

                        # directly unmount any filesystems for this container
                        for mount in get_mounts():
                            if mount.spec in loop_devices:
                                loop = loop_devices[mount.spec]
                                self.log.info('Unmounting %s..', mount.file)
                                mount.unmount()
                                loop.detach()
                            elif instance_path in Path(mount.file).parents:
                                self.log.info('Unmounting %s..', mount.file)
                                mount.unmount()
                        rmtree(str(instance_path), ignore_errors=True)

    def cleanup_posix_users(self):
        remove_users(self.node_state.unused_app_user_names)

    def cleanup_instances(self):
        running_instance_names = set([inst.name for inst in self.node_state.instances])
        for instance in self.cluster_state[self.host].instances:
            lock_file = LockFile(str(LOCK_PATH / instance['instance_name']))
            # TODO testable method for this condition
            if instance['lifecycle_phase'] not in ('START', 'KILL')\
                    and not lock_file.is_locked()\
                    and instance['instance_name'] not in running_instance_names:
                self.ctrl.delete_instance(instance['instance_name'])

    def cleanup(self):
        self.log.info('Running cleanup..')

        self.cleanup_instances()

        # remove stale symlinks of died applications
        for path in NETNS_PATH.iterdir():
            # hack because .resolve was crashing
            if not os.path.exists(os.readlink(str(path))):
                path.unlink()

        self.cleanup_posix_users()
        self.cleanup_containers()

    def reboot(self):
        try:
            for instance in self.node_state.instances:
                self.stop_application_instance(instance)
        except:
            self.log.exception('Error while stopping instances, rebooting any way..')
        self.log.info('Rebooting..')
        reboot()
        exit(0)  # stop the agent

    @coroutine
    def haproxy_loop(self):
        while True:
            now = time()
            loop_wait = now - self._last_haproxy_loop
            self._last_haproxy_loop = now
            try:
                self.log.debug('Starting HAProxy management loop (waited %d s)', loop_wait)
                self.cluster_state = self.ctrl.poll()
                self.update_and_cleanup_haproxies()
            except:
                self.log.exception('Execution failure in haproxy management loop')
            finally:
                if self.test_mode:
                    return
                yield from sleep(MAIN_LOOP_INTERVAL)

    @coroutine
    def instance_loop(self):
        while True:
            try:
                yield from self.instance_iteration()
            except:
                self.log.exception('Execution failure in instance management loop')
            finally:
                if self.test_mode:
                    return
                yield from sleep(MAIN_LOOP_INTERVAL)

    @coroutine
    def instance_iteration(self):
        self.log.debug('Starting instance management loop')
        self.setproctitle('polling..')
        self.log.debug('Getting node %s from Core Manager...', self.node_state.node_name)
        this_node = self.core_manager.get_node(self.node_state.node_name)

        self.log.debug('Checking Flags')
        if this_node.reboot:
            self.reboot()

        if this_node.maintenance:
            self.log.info('In maintenance')

        self.log.debug('Checking Cluster State')
        self.cluster_state = self.ctrl.poll()
        # if self.host not in self.cluster_state:
        #    self.log.info('This node is not part of the current cluster state, re-registering..')
        #    self.register()
        #    return

        self.log.debug('Calculating changes')
        target_instances = self.cluster_state[self.host].instances
        stop_list, start_list, clear_list = self.node_state.diff(target_instances)

        self.log.debug('Removing %d non-matching application instances..', len(stop_list))
        for instance_name in clear_list:
            self.ctrl.delete_instance(instance_name)

        self.log.debug('Stopping %d application instances..', len(stop_list))
        for instance in stop_list:
            self.stop_application_instance(instance)

        self.log.debug('Starting %d application instances..', len(start_list))
        tasks = [async(self.start_application_instance(instance)) for instance in start_list]
        if tasks and not this_node.maintenance:
            # applications are only started if the node is not in maintenance mode
            yield from async(wait(tasks))

    @coroutine
    def iptables_loop(self):
        while True:
            now = time()
            loop_wait = now - self._last_iptable_loop
            self._last_iptable_loop = now
            try:
                self.log.debug('Starting iptables management loop (waited %d s)', loop_wait)
                self.cluster_state = self.ctrl.poll()
                tasks = [async(self.open_container(instance)) for instance in self.node_state.instances]
                if tasks:
                    yield from wait(tasks)
            except:
                self.log.exception('Execution failure in iptables management loop')
            finally:
                if self.test_mode:
                    return
                yield from sleep(MAIN_LOOP_INTERVAL)

    @coroutine
    def cleanup_loop(self):
        while True:
            now = time()
            loop_wait = now - self._last_cleanup_loop
            self._last_cleanup_loop = now
            try:
                self.log.debug('Starting Cleanup loop (waited %d s)', loop_wait)
                self.cluster_state = self.ctrl.poll()
                self.setproctitle('cleaning up..')
                self.cleanup()
            except:
                self.log.exception('Execution failure in cleanup loop')
            finally:
                if self.test_mode:
                    return
                yield from sleep(CLEANUP_INTERVAL)

    def main_loop(self):
        """
        Main execution loop
        """
        loop = get_event_loop()
        try:
            tasks = [async(self.haproxy_loop()), async(self.instance_loop()), self.cleanup_loop(), self.iptables_loop()]
            loop.run_until_complete(wait(tasks))
        except (KeyboardInterrupt, SystemExit, EOFError):
            loop.close()
            raise

    def register(self):
        self.log.info('Registering %s..', self.host)
        self.ctrl.register(self.node_state.host, self.node_state.public_dict)
        this_node = self.core_manager.get_node(self.node_state.node_name)
        try:
            this_node.reset_reboot_flag()
        except Exception as e:
            self.log.warning('Failed to reset reboot flag in Core Manager: %s', e)

    def stop_application_instance(self, instance: Instance):
        self.log.info('Stopping application instance %s..', instance.name)
        self.setproctitle('stopping {}..', instance.name)
        instance.lifecycle_phase = 'KILLING'
        try:
            self.ctrl.update_instance(instance)
        except Exception as e:
            self.log.warning('Failed to update instance %s to %s in controller: %s', instance.name,
                             instance.lifecycle_phase, e)
        self.docker.stop(instance.container_id, timeout=instance.manifest.shutdown_timeout_seconds)
        self.ctrl.delete_instance(instance.name)

    @coroutine
    def start_application_instance(self, instance_request):
        """
        Starts missing application instances
        Pulls application image and application manifest from Application Repository
        Creates local POSIX user (e.g. using application instance ID [UUID])
        Creates Docker container with instance ID in its name
        Configures resource limits (cgroups memory, cpu, ulimits, ..)
        Configures network rules according to application manifest
        Configures service routing (HAProxy) according to application manifest
        Configures DNS for service discovery
        Configures storage volumes according to application manifest
        Configures Log Collector with desired log file formats according to application manifest
        Starts Docker container (with Zalando AppArmor profile as previously generates POSIX user)
        """

        self.log.info('Starting application instance %s (%s/%s:%s)..', instance_request['instance_name'],
                      instance_request['repository_name'], instance_request['application_name'],
                      instance_request['application_version'])

        try:
            manifest = yield from run_async(self.registry.get_manifest,
                                            instance_request['repository_name'],
                                            instance_request['application_name'],
                                            instance_request['application_version'])
            instance = StartingInstance(instance_request, manifest)

            self.setproctitle('starting {}..', instance.name)

            with LockFile(str(LOCK_PATH / instance.name)):

                # TODO yield from
                self.ctrl.update_instance(instance)

                starter = Starter(instance, self.ippool, self.cluster_state, location=self.my_location,
                                  expose_services=self.expose_services)
                yield from async(starter.start())

                for inst in self.node_state.instances:
                    if inst.name == instance.name:
                        # TODO yield from
                        self.ctrl.update_instance(inst)
        except Exception as e:
            if isinstance(e, StartupError):
                self.log.error('Failed to start application instance %s: %s', instance_request['instance_name'], e)
            else:
                self.log.exception('Failed to start application instance %s', instance_request['instance_name'])
            self.ctrl.delete_instance(instance_request['instance_name'])

    def setup(self):
        """
        Initial environment setup, e.g. create required directories
        """
        for directory in NETNS_PATH, CACHE_PATH, LOCK_PATH, DATA_PATH, PERSISTENCE_PATH, INSTANCE_DATA_PATH:
            try:
                directory.mkdir(parents=True)
            except FileExistsError:
                # the dir already exists great
                pass

        self.ippool = IPPool.autoconfigure()
        self.setup_bridge()
        self.setup_routes()
        self.check_environment()

    def setup_bridge(self):
        out = execute('brctl show {bridge}', False, bridge=BRIDGE_NAME)
        if 'No such device' in out:
            gateways = netifaces.gateways()['default']
            execute('brctl addbr {bridge}', bridge=BRIDGE_NAME)
            execute('ip link set dev {bridge} up', bridge=BRIDGE_NAME)
            iface = self.ippool.iface
            execute('brctl addif {bridge} {interface}', bridge=BRIDGE_NAME, interface=iface)

            addresses = netifaces.ifaddresses(iface)
            inet = addresses.get(netifaces.AF_INET)
            inet6 = addresses.get(netifaces.AF_INET6)
            for a in inet + inet6:
                if '%' not in a['addr']:
                    net = netaddr.IPNetwork('{}/{}'.format(a['addr'], a['netmask']))
                    execute('ip addr add {network} dev {bridge}', network=net, bridge=BRIDGE_NAME)
                    execute('ip addr del {network} dev {device}', network=net, device=iface)
            # wait for IPv6 duplicate address detection
            wait_for_dad(BRIDGE_NAME)
            # add default gateways back to bridge device:
            gw = gateways.get(netifaces.AF_INET)
            if gw and gw[1] == iface:
                execute('ip route add default via {nexthop} dev {bridge}', nexthop=gw[0], bridge=BRIDGE_NAME)
            gw = gateways.get(netifaces.AF_INET6)
            if gw and gw[1] == iface:
                # note sure why, but we should remove the old default route
                execute('ip -6 route del default via {nexthop} dev {interface}', nexthop=gw[0], interface=iface)
                execute('ip -6 route add default via {nexthop} dev {bridge}', nexthop=gw[0], bridge=BRIDGE_NAME)
            else:
                self.log.warning('No IPv6 default gateway found: %s', gw)

    @staticmethod
    def _find_my_network(networks: list, my_net: IPNetwork) -> dict:
        for loc in networks:
            # find our own location
            net = IPNetwork('{}/{}'.format(loc['ipv6_subnet_ip'], loc['ipv6_subnet_prefixlen']))
            if net == my_net:
                return loc

    def setup_routes(self):
        self.log.info('Getting locations to setup routes..')
        networks = self.core_manager.get_networks()
        my_net = self.ippool.net
        self.log.debug('My IPv6 network is %s', my_net)
        loc = self._find_my_network(networks, my_net)
        self.my_location = loc
        if loc:
            self.log.info('Location is %s, default gateway is %s, nameserver is %s',
                          loc['location_name'], loc['ipv6_gateway'], loc['ipv6_dns'])
            execute('ip -6 route add default via {ipv6_gateway}', **loc)
            setup_routes(loc['ipv6_routes'])
            with RESOLV_CONF_PATH.open('w', encoding='ascii') as fd:
                fd.write('nameserver {}\n'.format(loc['ipv6_dns']))
        else:
            self.log.warning('Could not find my own location by IPv6 subnet (%s)', my_net)

    def check_environment(self):
        hostname = 'zalando.de'
        try:
            get_host_ip(hostname)
        except:
            self.log.warning('Could not resolve IPv6 address of %s. DNS/DNS64 seems not to work.', hostname)

    @coroutine
    def open_container(self, instance):
        """
        Allow instances that need a service to access containers that have that service
        """
        self.log.debug('Opening Container Network for %s', instance.name)

        manifest = yield from run_async(self.registry.get_manifest, instance.repository_name, instance.application_name,
                                        instance.application_version)
        if not manifest:
            self.log.warning('Could not find manifest for running application instance %s', instance.name)
            return
        provided_services = [service for service in manifest.services if manifest.data and service['type'] == 'provide']

        self.log.debug('%s provides %d service(s)', instance.name, len(provided_services))

        for service in provided_services:
            service_name = service['service_name']
            service_version = service['service_version']
            self.log.debug('Configuring iptables for %s:%s in %s', service_name, service_version, instance.name)

            # Accept connections from the load balancer if needed
            if self.lb_updater:
                self.log.debug('Finding %s:%s in %s on load balancer', service_name, service_version, instance.name)
                pools = yield from run_async(self.lb_updater.discover_in_load_balancer, service_name, service_version)
            else:
                pools = []

            for pool in pools:
                self.log.debug('Opening %s to %s', instance.name, pool)
                with Namespace(instance.pid, 'net'):
                    # FIXME
                    ip = '::/0'
                    ensure('INPUT -p tcp --dport {port} -s {ip} -j ACCEPT', port=service['local_port'], ip=ip)
                    ensure('OUTPUT -p tcp --sport {port} -d {ip} -j ACCEPT', port=service['local_port'], ip=ip)

            # Accept connections from other applications
            source_ips = self.cluster_state.discover_dependents(self.registry, instance.zone_name, service_name,
                                                                service_version)
            with Namespace(instance.pid, 'net'):
                for ip in source_ips:
                    # allow connections from source IP to our provided service
                    ensure('INPUT -p tcp --dport {port} -s {ip} -j ACCEPT', port=service['local_port'], ip=ip)
                    # allow responding from our provided service to the source IP
                    ensure('OUTPUT -p tcp --sport {port} -d {ip} -j ACCEPT', port=service['local_port'], ip=ip)

    def update_and_cleanup_haproxies(self):
        """
        Updates the endpoints of each haproxy and stops haproxy instances that aren't needed anymore
        """
        self.log.debug('Updating HAProxies..')
        for container_id, proxies in self.haproxy_manager.proxies.items():
            for proxy in proxies:
                still_needed = self.docker.container_is_running(container_id)
                if still_needed:
                    for service in proxy.services:
                        endpoints = self.cluster_state.discover_service(self.registry, self.core_manager,
                                                                        proxy.app_instance.zone_name, service.name,
                                                                        service.version)
                        self.allow_haproxy(container_id, endpoints)
                        # Update service endpoints
                        service.endpoints = endpoints

                    if proxy.restart():
                        self.haproxy_manager.register(proxy)  # persist changes
                else:
                    self.haproxy_manager.remove(proxy)

    def allow_haproxy(self, container_id, endpoints):
        """
        Updates iptables
        """
        instance = self.node_state.get_instance_by_container_id(container_id)
        self.log.debug('Updating iptables for %s\'s HAProxies..', instance.name)
        with Namespace(instance.pid, 'net'):
            for endpoint in endpoints:
                # allow HAProxy to connect to remote (required) service
                ensure('OUTPUT -p tcp --dport {port} -d {ip} -j ACCEPT', port=endpoint.port, ip=endpoint.ip)
                # allow remote (required) service to answer HAProxy
                ensure('INPUT -p tcp --sport {port} -s {ip} -j ACCEPT', port=endpoint.port, ip=endpoint.ip)


def parse_core_seeds(arg: str) -> list:
    """
    Parse --core-seeds command line parameter which can either be a file name (e.g. /proc/cmdline) or a list of IPs
    """
    if arg.startswith('/'):
        core_seeds = None
        with open(arg) as fd:
            for line in fd:
                # parse kernel commandline parameters
                for word in line.split(None):
                    if word.startswith(CORE_SEEDS_ARG):
                        core_seeds = word[len(CORE_SEEDS_ARG):]
        if not core_seeds:
            raise ConfigurationError('Core seeds argument not found in {}'.format(arg))
    else:
        core_seeds = arg

    return core_seeds.split(',')


def setup_logging(loglevel, test_mode):
    logger = logging.getLogger()
    logger.setLevel(loglevel or logging.INFO)
    if test_mode:
        # we only need a short time format when running locally
        handler = ColorLoggingHandler(sys.stderr, datefmt='%H:%M:%S')
    else:
        handler = ColorLoggingHandler(sys.stderr)
    logger.addHandler(handler)
    logging.getLogger('urllib3.connectionpool').setLevel(logging.WARNING)
    logging.getLogger('requests.packages.urllib3.connectionpool').setLevel(logging.WARNING)
    logging.getLogger('asyncio').setLevel(logging.WARNING)


def check(msg: str, assertion: bool):
    """
    Helper function to raise a PreflightError if the assertion failed

    >>> check('ok', True)
    >>> check('error', False)
    Traceback (most recent call last):
      ...
    pequod_agent.errors.PreflightError: Preflight check failed: error
    """
    if not assertion:
        raise PreflightError(msg)


def preflight_checks():
    """
    Check that all requirements are met before doing anything real
    """
    check('The agent must be run as root', os.getuid() == 0)


def parse_args(args):
    """

    >>> parse_args(['--bootstrap', 'a', 'b']).bootstrap
    ['a', 'b']
    """
    parser = ArgumentParser()
    group = parser.add_mutually_exclusive_group()
    group.add_argument('-v', '--verbose', help='Verbose (debug) logging', action='store_const', const=logging.DEBUG,
                       dest='loglevel')
    group.add_argument('-q', '--quiet', help='Silent mode, only log warnings', action='store_const',
                       const=logging.WARN, dest='loglevel')
    group = parser.add_mutually_exclusive_group()
    group.add_argument('-s', '--core-seeds', help='IP(s) of core nodes (hosting Cluster Controller) or "/proc/cmdline"',
                       type=parse_core_seeds)
    group.add_argument('-b', '--bootstrap', nargs='+', help='Bootstrap from given Docker images or YAML file(s)')
    group.add_argument('-i', '--inspect', nargs='*', help='Inspect agent environment (list containers etc) and exit',
                       metavar='QUERY')
    parser.add_argument('-t', '--test', help='Test run, only run the main loop once', action='store_true')
    parser.add_argument('--expose-services', help='Automatically expose TCP ports of all provided services',
                        action='store_true')
    group.add_argument('-V', '--version', help='Print the agent version number and exit', action='store_true')
    return parser.parse_args(args)


def main():
    args = parse_args(sys.argv[1:])

    if args.version:
        print('Pequod Cluster Agent', pequod_agent.__version__)
        exit(0)

    setup_logging(args.loglevel, args.test)

    try:
        preflight_checks()
        node_state = NodeState()
        if args.bootstrap:
            cluster = Cluster.setup_bootstrap(node_state.host, args.bootstrap)
        elif args.core_seeds:
            cluster = Cluster.setup_core_seeds(node_state.host, args.core_seeds)
        elif args.inspect is not None:
            inspect(node_state, args.inspect)
            exit(0)
        else:
            raise NotImplementedError('Please specify either "--bootstrap", "--core-seeds" or "--inspect"')
        agent = Agent(cluster, node_state, test_mode=args.test, expose_services=args.expose_services)
        agent.setup()
        agent.register()
        agent.main_loop()
    except (ConfigurationError, PreflightError) as e:
        logging.exception('%s', e)
        exit(126)
    except (KeyboardInterrupt, EOFError):
        print('\nStopped by the user')
    except (SystemExit):
        pass
    except Exception as e:
        logging.exception('Fatal error: %s', e)
        exit(1)


if __name__ == '__main__':
    main()
