#!/usr/bin/env python3
from pequod_agent.models.application import parse_application_spec
from pequod_agent.utils import generate_random_name

__author__ = 'hjacobs'

import yaml
import logging
from pequod_agent.cluster.controllers.base import Controller
from pequod_agent.docker_manager import DockerManager
from pequod_agent.models.cluster_state import ClusterState


class BootstrapController(Controller):
    """
    Controller to run in standalone mode to bootstrap management nodes
    """
    def __init__(self, host, bootstrap_files):
        self.log = logging.getLogger('pequod.controller')
        docker = DockerManager()
        docker = DockerManager()
        self.hostname = host
        self.instances = []
        for bootstrap_file in bootstrap_files:
            try:
                app = parse_application_spec(bootstrap_file)
            except ValueError:
                app = None
            if app:
                manifest = docker.get_manifest_from_image(bootstrap_file)
                resources = {}
                for res, val in manifest.resources.items():
                    for k, v in val.items():
                        if k.startswith('default_'):
                            resources['{}_{}'.format(res, k.split('_')[1])] = v
                name_prefix = 'bootstrap-{}-{}-{}-'.format(app.repository_name,
                                                           app.application_name,
                                                           app.application_version)[:24]
                instance_name = generate_random_name(name_prefix, 4)
                self.instances.append({
                    'instance_name': instance_name,
                    'repository_name': app.repository_name,
                    'application_name': app.application_name,
                    'application_version': app.application_version,
                    'node_name': host,
                    'lifecycle_phase': 'START',
                    'zone_name': 'bootstrap',
                    'resources': resources

                })
            else:
                with open(bootstrap_file) as fd:
                    data = yaml.safe_load(fd)
                instances = data.get('instances', [])
                if not isinstance(instances, list):
                    raise ValueError('Invalid instances in bootstrap file')
                for inst in instances:
                    if not isinstance(inst['instance_name'], str):
                        raise ValueError('Instance name must be a string')
                    inst['lifecycle_phase'] = 'START'
                    inst['node_name'] = host
                    if 'zone_name' not in inst:
                        inst['zone_name'] = 'bootstrap'
                self.instances += instances

    def register(self, node_host: str, node_state: dict):
        self.log.debug('Register: %s', node_state)
        for running_inst in node_state['instances']:
            for inst in self.instances:
                if running_inst['instance_name'] == inst['instance_name']:
                    inst['ipv6'] = running_inst['metadata']['ipv6']

    def update_node(self, node_state):
        self.log.debug('Update: %s', node_state.public_dict)

    def update_instance(self, instance):
        self.log.debug('Update: %s', instance.public_dict)
        for inst in self.instances:
            if instance.name == inst['instance_name']:
                inst['ipv6'] = instance.ip_address

    def delete_instance(self, instance_name):
        self.log.debug('Delete: %s', instance_name)

    def poll(self):
        self.log.debug('Poll: %s', self.instances)
        return ClusterState({self.hostname: {'instances': self.instances}})
