#!/usr/bin/env python3
from collections import Mapping
import logging
from pequod_agent.cluster.manager.base import CoreManager
from pequod_agent.cluster.service.app import INFRASTRUCTURE_ZONE
from pequod_agent.constants.ports import CONTROLLER_PORT, MANAGER_PORT, REGISTRY_PORT, CASSANDRA_PORT, REPLICATOR_PORT, \
    AUTH_PORT
from pequod_agent.models.version_range import version_matches
from pequod_agent.networking.ip import get_host_ip
from pequod_agent.networking.routing.service import ServiceEndpoint
from pequod_agent.cluster.registries.base import Registry

CORE_SERVICES = {
    ('pequod-core-auth', '1'): AUTH_PORT,
    ('pequod-core-replicator', '1'): REPLICATOR_PORT,
    ('pequod-core-manager', '1'): MANAGER_PORT,
    ('pequod-core-registry', '1'): REGISTRY_PORT,
    ('pequod-core-controller', '1'): CONTROLLER_PORT,
    ('pequod-core-cassandra', '1'): CASSANDRA_PORT,
}


class Node:
    def __init__(self, hostname, state):
        self._state = state
        self.hostname = hostname

    @property
    def instances(self) -> list:
        return self._state['instances']


class ClusterState(Mapping):
    def __init__(self, state: dict, core_seeds=None):
        self.log = logging.getLogger('pequod.cluster-state')
        self._state = state
        self.core_seeds = core_seeds

    def __getitem__(self, node_hostname: str):
        return Node(node_hostname, self._state[node_hostname])

    def __iter__(self):
        for node_hostname in self._state:
            yield node_hostname

    def __len__(self):
        return len(self._state)

    def discover_core_endpoints(self, zone_name: str, service_name: str, version_range: str):
        """
        resolve hard-coded core services (any application should be able to use them)
        """
        if self.core_seeds and zone_name == INFRASTRUCTURE_ZONE:
            port = CORE_SERVICES.get((service_name, version_range))
            if port:
                for ip in sorted(self.core_seeds):
                    ip = get_host_ip(ip, port)
                    yield ServiceEndpoint(ip, port)

    def discover_external_endpoints(self, manager: CoreManager, zone_name: str, service_name: str, version_range: str):
        """
        resolve external services defined in cluster config service
        """
        yield from manager.discover_external_service(zone_name, service_name, version_range)

    def discover_instance_endpoints(self,
                                    registry: Registry, zone_name: str, service_name: str, version_range: str) -> list:
        """
        find all service endpoints provided by running application instances in the same zone
        """

        for node in self.values():
            for instance in node.instances:
                if instance['zone_name'] != zone_name:
                    # make sure we only consider instances in same zone
                    continue
                if instance['lifecycle_phase'] != 'RUNNING':
                    # make sure we only consider running instances
                    continue
                if not instance.get('ipv6'):
                    self.log.debug('Instance %s does not have an IPv6 address', instance['instance_name'])
                    continue
                manifest = registry.get_manifest(instance['repository_name'],
                                                 instance['application_name'],
                                                 instance['application_version'])
                if not manifest:
                    continue
                self.log.debug('Found manifest: %s', manifest)
                provided_services = [service for service in manifest.services if service['type'] == 'provide']
                for service in provided_services:
                    in_version_range = version_matches(service['service_version'], version_range)
                    if service['service_name'] == service_name and in_version_range:
                        self.log.debug('%s', instance)
                        yield ServiceEndpoint(instance['ipv6'], service['local_port'])

    def discover_service(self,
                         registry: Registry, manager, zone_name: str, service_name: str, version_range: str) -> list:
        self.log.info('Trying to discover endpoints for required service %s:%s in %s zone..',
                      service_name, version_range, zone_name)

        endpoints = list(self.discover_core_endpoints(zone_name, service_name, version_range))

        if not endpoints:
            endpoints = list(self.discover_external_endpoints(manager, zone_name, service_name, version_range))
            endpoints.extend(list(self.discover_instance_endpoints(registry, zone_name, service_name, version_range)))

        self.log.debug('endpoints: %s', [str(e) for e in endpoints])

        return endpoints

    def discover_dependents(self, registry: Registry, zone_name: str, service_name: str, version_range: str) -> list:
        self.log.info('Trying to discover instances that require service %s:%s in %s zone..',
                      service_name, version_range, zone_name)
        source_ips = []

        for node in self.values():
            for instance in node.instances:
                if instance['zone_name'] != zone_name:
                    # make sure we only consider instances in same zone
                    continue
                manifest = registry.get_manifest(instance['repository_name'], instance['application_name'],
                                                 instance['application_version'])
                if not manifest:
                    continue
                self.log.debug('Found manifest: %s', manifest)
                required_services = [service for service in manifest.services if service['type'] == 'require']
                for service in required_services:
                    in_version_range = version_matches(service['service_version'], version_range)
                    if service['service_name'] == service_name and in_version_range:
                        self.log.debug('%s', instance)
                        # IPv6 is not yet set when instance is in STARTING state
                        if instance.get('ipv6'):
                            source_ips.append(instance['ipv6'])

        self.log.debug('source IPs: %s', source_ips)

        return source_ips
