import operator
import sys

VERSION_MIN = ()
VERSION_MAX = tuple([sys.maxsize] * 6)


def int_or_str(s: str):
    try:
        return int(s)
    except:
        return 0


def version_matches(version: str, version_range: str) -> bool:
    """
    Check whether the given version matches the given version range.

    :param version: The given version string.
    :param version_range: Version range specification as string.
    :return: True if version is in range, False otherwise.

    >>> version_matches('1', '1')
    True

    >>> version_matches('2', '[1,]')
    True

    >>> version_matches('2', '(2,]')
    False
    """
    return version in VersionRange.parse(version_range)


class VersionRange:

    def __init__(self, interval, interval_left_open=False, interval_right_open=False):
        self.interval = interval
        self.interval_left_open = interval_left_open
        self.interval_right_open = interval_right_open

    @staticmethod
    def parse_version(ver: str) -> tuple:
        """
        Parse a single version string and return the canonical version tuple.

        >>> VersionRange.parse_version('1')
        (1,)

        >>> VersionRange.parse_version('a')
        (0,)

        >>> VersionRange.parse_version('0.1')
        (0, 1)
        """
        return tuple(map(int_or_str, ver.split('.')))

    @classmethod
    def parse(cls, spec: str):
        """
        Parse a version range specification.

        :param spec: version range string
        :return: a VersionRange instance
        """

        if (spec.startswith('[') or spec.startswith('(')) and ',' in spec:
            a, b = spec.strip('[()]').split(',', 1)
            if a:
                a = cls.parse_version(a)
            else:
                a = VERSION_MIN
            if b:
                b = cls.parse_version(b)
            else:
                b = VERSION_MAX

            return cls((a, b), interval_left_open=spec[0] == '(', interval_right_open=spec[-1] == ')')
        else:
            return cls((cls.parse_version(spec), cls.parse_version(spec)))

    def __contains__(self, version: str) -> bool:
        ver = self.parse_version(version)

        op1 = operator.lt if self.interval_left_open else operator.le
        op2 = operator.lt if self.interval_right_open else operator.le
        if op1(self.interval[0], ver) and op2(ver, self.interval[1]):
                return True
        return False
